import logging
import shutil
import tempfile
from pathlib import Path
from urllib.parse import urlparse

import click
import livereload
from livereload.handlers import LiveReloadHandler
from rich.logging import RichHandler

from .config import Config
from .constants import (
    DEFAULT_OUTPUT_DIR,
    EXPECTED_CONFIG_LOCATION,
    HIGHLIGHTJS_THEMES_VERSION,
    REVEALJS_VERSION,
    VERSION,
)
from .markupgenerator import MarkupGenerator

logger = logging.getLogger()
logger.setLevel("DEBUG")
logger.addHandler(RichHandler(show_path=False))

LiveReloadHandler.DEFAULT_RELOAD_TIME = (
    0  # https://github.com/lepture/python-livereload/pull/244
)

################################################################################

context_settings = {"help_option_names": ["-h", "--help"], "max_content_width": 120}

files_argument_data = {
    "metavar": "FILENAME|PATH",
}

config_file_argument_data = {
    "metavar": "FILENAME",
    "default": EXPECTED_CONFIG_LOCATION,
    "help": "Provide a specific MkSlides-Reveal config file.",
}


@click.group(context_settings=context_settings)
@click.version_option(
    VERSION,
    "-V",
    "--version",
    message=f"mkslides, version {VERSION}\nreveal.js, version {REVEALJS_VERSION}\nhighlight.js themes, version {HIGHLIGHTJS_THEMES_VERSION}"
    "",
)
def cli() -> None:
    """MkSlides - Slides with Markdown using the power of Reveal.js."""


def read_config(config_location: str) -> Config:
    config_path = Path(config_location).resolve()
    config = Config()

    if config_path.exists():
        logger.info(f'Config file found at "{config_path.absolute()}"')
        config.merge_config_from_file(config_path)

    return config


def parse_ip_port(
    ip_port_str: str,
) -> tuple[str, int]:
    urlparse_result = urlparse(f"//{ip_port_str}")
    ip = urlparse_result.hostname
    port = urlparse_result.port

    assert ip, f"Invalid IP address: {ip_port_str}"
    assert port, f"Invalid port: {ip_port_str}"

    return ip, port


def generate(config_file, input_path: Path, output_directory: Path) -> None:
    config = read_config(config_file)
    markup_generator = MarkupGenerator(config, output_directory)
    markup_generator.create_output_directory()
    markup_generator.process_markdown(input_path)


@cli.command()
@click.argument("files", **files_argument_data)
@click.option("-f", "--config-file", **config_file_argument_data)
@click.option(
    "-d",
    "--site-dir",
    help="The directory to output the result of the slides build.",
    metavar="PATH",
    default=DEFAULT_OUTPUT_DIR,
)
def build(files, config_file, site_dir) -> None:
    """Build the MkDocs documentation.

    FILENAME|PATH is the path to the Markdown file, or the directory containing Markdown files.
    """
    logger.info("Command: build")

    input_path = Path(files).resolve(strict=True)
    output_directory = Path(site_dir).resolve(strict=False)

    generate(config_file, input_path, output_directory)


@cli.command()
@click.argument("files", **files_argument_data)
@click.option(
    "-a",
    "--dev-addr",
    help="IP address and port to serve slides locally.",
    metavar="<IP:PORT>",
    default="localhost:8000",
)
@click.option(
    "-o",
    "--open",
    "open_in_browser",
    help="Open the website in a Web browser after the initial build finishes.",
    is_flag=True,
)
@click.option(
    "--watch-index-theme",
    help="Include the index theme in list of files to watch for live reloading.",
    is_flag=True,
)
@click.option(
    "--watch-index-template",
    help="Include the index template in list of files to watch for live reloading.",
    is_flag=True,
)
@click.option(
    "--watch-slides-theme",
    help="Include the slides theme in list of files to watch for live reloading.",
    is_flag=True,
)
@click.option(
    "--watch-slides-template",
    help="Include the slides template in list of files to watch for live reloading.",
    is_flag=True,
)
@click.option("-f", "--config-file", **config_file_argument_data)
def serve(
    files,
    dev_addr,
    open_in_browser,
    watch_index_theme,
    watch_index_template,
    watch_slides_theme,
    watch_slides_template,
    config_file,
) -> None:
    """Run the builtin development server.

    FILENAME|PATH is the path to the Markdown file, or the directory containing Markdown files.
    """
    logger.info("Command: serve")

    input_path = Path(files).resolve(strict=True)
    site_dir = tempfile.mkdtemp(prefix="mkslides_")
    output_directory = Path(site_dir).resolve(strict=False)

    generate(config_file, input_path, output_directory)
    config = read_config(config_file)

    # Livereload

    def reload() -> None:
        logger.info("Reloading ...")
        generate(config_file, input_path, output_directory)

    try:
        server = livereload.Server()
        server._setup_logging = (
            lambda: None
        )  # https://github.com/lepture/python-livereload/issues/232

        watched_paths = [
            files,
            config_file,  # TODO: reload config
        ]

        if watch_index_theme:
            watched_paths.append(config.get_index_theme())
        if watch_index_template:
            watched_paths.append(config.get_index_template())
        if watch_slides_theme:
            watched_paths.append(config.get_slides_theme())
        if watch_slides_template:
            watched_paths.append(config.get_slides_template())

        for path in watched_paths:
            if path:
                path = Path(path).resolve(strict=True)
                logger.info(f'Watching: "{path.absolute()}"')
                server.watch(filepath=path.absolute().as_posix(), func=reload, delay=1)

        ip, port = parse_ip_port(dev_addr)

        server.serve(
            host=ip,
            port=port,
            root=output_directory,
            open_url_delay=0 if open_in_browser else None,
        )

    finally:
        if output_directory.exists():
            shutil.rmtree(output_directory)
            logger.info(f'Removed "{output_directory}"')


if __name__ == "__main__":
    cli()



import json
import re
from importlib import metadata, resources

from jinja2 import Environment, FileSystemLoader, PackageLoader, select_autoescape

MD_LINK_REGEX = re.compile(
    r"""
    \[.*?\]                     # Alt text
    \((?P<location>[^()<>]*?)\) # Location without angle brackets
    """,
    re.VERBOSE,
)

MD_ESCAPED_LINK_REGEX = re.compile(
    r"""
    \[.*?\]                     # Alt text
    \(<(?P<location>[^<>]*?)>\) # Location with mandatory angle brackets
    """,
    re.VERBOSE,
)

HTML_IMAGE_REGEX = re.compile(
    r"""
    <img                        # Start of the image
    .+?                         # Any attributes
    src=                        # src attribute
        (?P<delimiter>['\"])    # Delimiter
        (?P<location>.+?)       # Image location
        (?P=delimiter)          # Delimiter
    .*?                         # Any attributes
    >                           # End of the image
    """,
    re.VERBOSE,
)

HTML_BACKGROUND_IMAGE_REGEX = re.compile(
    r"""
    <!--                        # Start of the comment
    .*?                         # Any content
    data-background-image=      # data-background-image attribute
        (?P<delimiter>['\"])    # Delimiter
        (?P<location>.+?)       # Image location
        (?P=delimiter)          # Delimiter
    .*?                         # Any content
    -->                         # End of the comment
    """,
    re.VERBOSE,
)

EXPECTED_CONFIG_LOCATION = "mkslides.yml"
DEFAULT_OUTPUT_DIR = "site"

ASSETS_RESOURCE = resources.files("assets")
DEFAULT_CONFIG_RESOURCE = ASSETS_RESOURCE.joinpath("mkslides.default.yml")
REVEALJS_RESOURCE = ASSETS_RESOURCE.joinpath("reveal.js")
REVEALJS_THEMES_RESOURCE = REVEALJS_RESOURCE.joinpath("dist", "theme")
HIGHLIGHTJS_RESOURCE = ASSETS_RESOURCE.joinpath("highlight.js")
HIGHLIGHTJS_THEMES_RESOURCE = HIGHLIGHTJS_RESOURCE.joinpath("build", "styles")

DEFAULT_JINJA2_ENVIRONMENT = Environment(
    loader=PackageLoader("assets"),
    autoescape=select_autoescape(),
)
DEFAULT_INDEX_TEMPLATE = DEFAULT_JINJA2_ENVIRONMENT.get_template("index.html.jinja")
DEFAULT_SLIDESHOW_TEMPLATE = DEFAULT_JINJA2_ENVIRONMENT.get_template(
    "slideshow.html.jinja",
)
LOCAL_JINJA2_ENVIRONMENT = Environment(loader=FileSystemLoader("."))

VERSION = metadata.version("mkslides")

REVEALJS_VERSION = None
with REVEALJS_RESOURCE.joinpath("package.json").open(encoding="utf-8-sig") as f:
    REVEALJS_VERSION = json.load(f)["version"]

HIGHLIGHTJS_THEMES_VERSION = None
with HIGHLIGHTJS_RESOURCE.joinpath("build", "package.json").open(
    encoding="utf-8-sig",
) as f:
    HIGHLIGHTJS_THEMES_VERSION = json.load(f)["version"]



import logging
from pathlib import Path

import yaml

from .constants import DEFAULT_CONFIG_RESOURCE

logger = logging.getLogger(__name__)


class Config:
    schema = {
        "type": "object",
        "properties": {
            "price": {"type": "number"},
            "name": {"type": "string"},
        },
    }

    def __init__(self) -> None:
        with DEFAULT_CONFIG_RESOURCE.open() as f:
            self.__config = yaml.safe_load(f)

        logger.info(f'Default config loaded from "{DEFAULT_CONFIG_RESOURCE}"')
        logger.info(f"Default config: {self.__config}")

    def get_index_title(self) -> str | None:
        return self.__get("index", "title")

    def get_index_favicon(self) -> str | None:
        return self.__get("index", "favicon")

    def get_index_theme(self) -> str | None:
        return self.__get("index", "theme")

    def get_index_template(self) -> str | None:
        return self.__get("index", "template")

    def get_slides_favicon(self) -> str | None:
        return self.__get("slides", "favicon")

    def get_slides_theme(self) -> str | None:
        return self.__get("slides", "theme")

    def get_slides_highlight_theme(self) -> str | None:
        return self.__get("slides", "highlight_theme")

    def get_slides_template(self) -> str | None:
        return self.__get("slides", "template")

    def get_slides_separator(self) -> str | None:
        return self.__get("slides", "separator")

    def get_slides_separator_vertical(self) -> str | None:
        return self.__get("slides", "separator_vertical")

    def get_slides_separator_notes(self) -> str | None:
        return self.__get("slides", "separator_notes")

    def get_slides_charset(self) -> str | None:
        return self.__get("slides", "charset")

    def get_revealjs_options(self) -> dict | None:
        return self.__get("revealjs")

    def get_plugins(self) -> list | None:
        return self.__get("plugins")

    def merge_config_from_file(self, config_path: Path) -> None:
        with config_path.open(encoding="utf-8-sig") as f:
            new_config = yaml.safe_load(f)

            self.__config = self.__recursive_merge(self.__config, new_config)

            logger.info(f'Config merged from "{config_path}"')
            logger.info(f"Config: {self.__config}")

    def merge_config_from_dict(self, new_config: dict) -> None:
        self.__config = self.__recursive_merge(self.__config, new_config)

        logger.info("Config merged from dict")
        logger.info(f"Config: {self.__config}")

    def __get(self, *keys) -> str | dict | list | None:
        current_value = self.__config
        for key in keys:
            if isinstance(current_value, dict) and key in current_value:
                current_value = current_value[key]
            else:
                return None
        return current_value

    def __recursive_merge(self, current, new) -> dict:
        if new:
            for key, value in new.items():
                if isinstance(value, dict):
                    current[key] = self.__recursive_merge(current.get(key, {}), value)
                else:
                    current[key] = value

        return current



import datetime
import logging
import shutil
import time
from emoji import emojize
from importlib import resources
from importlib.resources.abc import Traversable
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import frontmatter
from natsort import natsorted

from .config import Config
from .constants import (
    DEFAULT_INDEX_TEMPLATE,
    DEFAULT_SLIDESHOW_TEMPLATE,
    HIGHLIGHTJS_THEMES_RESOURCE,
    HTML_BACKGROUND_IMAGE_REGEX,
    HTML_IMAGE_REGEX,
    LOCAL_JINJA2_ENVIRONMENT,
    MD_ESCAPED_LINK_REGEX,
    MD_LINK_REGEX,
    REVEALJS_RESOURCE,
    REVEALJS_THEMES_RESOURCE,
)

logger = logging.getLogger(__name__)


class MarkupGenerator:
    def __init__(
        self,
        config: Config,
        output_directory_path: Path,
    ) -> None:
        # Config

        self.config = config

        # Paths

        self.output_directory_path = output_directory_path.resolve(strict=False)
        logger.info(
            f'Requested output directory: "{self.output_directory_path.absolute()}"',
        )

        self.output_assets_path = self.output_directory_path / "assets"
        self.output_revealjs_path = self.output_assets_path / "reveal-js"

    def clear_output_directory(self) -> None:
        for item in self.output_directory_path.iterdir():
            if item.is_file():
                item.unlink()
            elif item.is_dir():
                shutil.rmtree(item)
        logger.info("Output directory cleared")

    def create_output_directory(self) -> None:
        if self.output_directory_path.exists():
            self.clear_output_directory()
        else:
            self.output_directory_path.mkdir(parents=True, exist_ok=True)
            logger.info("Output directory created")

        with resources.as_file(REVEALJS_RESOURCE) as revealjs_path:
            self.__copy(revealjs_path, self.output_revealjs_path)

    def process_markdown(self, input_path: Path) -> None:
        logger.info("Processing markdown")
        start_time = time.perf_counter()

        if input_path.is_dir():
            self.__process_markdown_directory(input_path)
        else:
            _, output_markup_path = self.__process_markdown_file(
                input_path,
                input_path.parent,
            )
            original_output_markup_path = output_markup_path

            if output_markup_path.stem != "index":
                output_markup_path.rename(output_markup_path.with_stem("index"))
                logger.info(
                    f'Renamed "{original_output_markup_path.absolute()}" to "{output_markup_path.absolute()}" as it was the only Markdown file',
                )

        end_time = time.perf_counter()
        logger.info(
            f"Finished processing markdown in {end_time - start_time:.2f} seconds",
        )

    ################################################################################

    def __process_markdown_file(
        self,
        md_file: Path,
        md_root_path: Path,
    ) -> tuple[dict, Path]:
        md_file = md_file.resolve(strict=True)
        md_root_path = md_root_path.resolve(strict=True)

        logger.info(f'Processing markdown file at "{md_file.absolute()}"')

        # Retrieve the frontmatter metadata and the markdown content

        content = md_file.read_text(encoding="utf-8-sig")
        metadata, markdown = frontmatter.parse(content)
        markdown = emojize(markdown, language="alias")

        # Get the relative path of reveal.js

        output_markup_path = self.output_directory_path / md_file.relative_to(
            md_root_path,
        )

        output_markup_path = output_markup_path.with_suffix(".html")
        relative_revealjs_path = self.output_revealjs_path.relative_to(
            output_markup_path.parent,
            walk_up=True,
        )

        revealjs_config = self.config.get_revealjs_options()

        # Copy the theme CSS

        relative_theme_path = None
        if theme := self.config.get_slides_theme():
            relative_theme_path = self.__copy_theme(
                output_markup_path,
                theme,
                REVEALJS_THEMES_RESOURCE,
            )

        # Copy the highlight CSS

        relative_highlight_theme_path = None
        if theme := self.config.get_slides_highlight_theme():
            relative_highlight_theme_path = self.__copy_theme(
                output_markup_path,
                theme,
                HIGHLIGHTJS_THEMES_RESOURCE,
            )

        # Copy the favicon

        relative_favicon_path = None
        if favicon := self.config.get_slides_favicon():
            relative_favicon_path = self.__copy_favicon(output_markup_path, favicon)

        # Retrieve the 3rd party plugins

        plugins = self.config.get_plugins()

        # Generate the markup from markdown

        # Refresh the templates here, so they have effect when live reloading
        slideshow_template = None
        if template_config := self.config.get_slides_template():
            slideshow_template = LOCAL_JINJA2_ENVIRONMENT.get_template(template_config)
        else:
            slideshow_template = DEFAULT_SLIDESHOW_TEMPLATE

        # https://revealjs.com/markdown/#external-markdown
        markdown_data_options = {}
        for key, value in {
            "data-separator": self.config.get_slides_separator(),
            "data-separator-vertical": self.config.get_slides_separator_vertical(),
            "data-separator-notes": self.config.get_slides_separator_notes(),
            "data-charset": self.config.get_slides_charset(),
        }.items():
            if value:
                markdown_data_options[key] = value

        markup = slideshow_template.render(
            favicon=relative_favicon_path,
            theme=relative_theme_path,
            highlight_theme=relative_highlight_theme_path,
            revealjs_path=relative_revealjs_path,
            markdown_data_options=markdown_data_options,
            markdown=markdown,
            revealjs_config=revealjs_config,
            plugins=plugins,
        )
        self.__create_file(output_markup_path, markup)

        # Copy local files

        self.__copy_local_files(md_file, md_root_path, markdown)

        return metadata, output_markup_path

    def __process_markdown_directory(self, md_root_path: Path) -> None:
        md_root_path = md_root_path.resolve(strict=True)
        logger.info(f"Processing markdown directory at {md_root_path.absolute()}")
        slideshows = []
        for md_file in md_root_path.glob("**/*.md"):
            (metadata, output_markup_path) = self.__process_markdown_file(
                md_file,
                md_root_path,
            )

            slideshows.append(
                {
                    "title": metadata.get("title", md_file.stem),
                    "location": output_markup_path.relative_to(
                        self.output_directory_path,
                    ),
                },
            )

        slideshows = natsorted(slideshows, key=lambda x: x["location"])

        logger.info("Generating index")

        index_path = self.output_directory_path / "index.html"

        # Copy the theme

        relative_theme_path = None
        if theme := self.config.get_index_theme():
            relative_theme_path = self.__copy_theme(index_path, theme)

        # Copy the favicon

        relative_favicon_path = None
        if favicon := self.config.get_index_favicon():
            relative_favicon_path = self.__copy_favicon(index_path, favicon)

        # Refresh the templates here, so they have effect when live reloading
        index_template = None
        if template_config := self.config.get_index_template():
            index_template = LOCAL_JINJA2_ENVIRONMENT.get_template(template_config)
        else:
            index_template = DEFAULT_INDEX_TEMPLATE

        content = index_template.render(
            favicon=relative_favicon_path,
            title=self.config.get_index_title(),
            theme=relative_theme_path,
            slideshows=slideshows,
            build_datetime=datetime.datetime.now(),
        )
        self.__create_file(index_path, content)

    def __copy_local_files(
        self,
        md_file: Path,
        md_root_path: Path,
        markdown: str,
    ) -> None:
        for regex in [
            MD_LINK_REGEX,
            MD_ESCAPED_LINK_REGEX,
            HTML_IMAGE_REGEX,
            HTML_BACKGROUND_IMAGE_REGEX,
        ]:
            for m in regex.finditer(markdown):
                location = m.group("location")

                if not self.__is_absolute_url(location):
                    image = Path(md_file.parent, location).resolve(strict=True)
                    self.__copy_to_output_relative_to_md_root(image, md_root_path)

    def __copy_theme(
        self,
        file_using_theme_path: Path,
        theme: str,
        default_theme_resource: Traversable | None = None,
    ) -> Path | str:
        if self.__is_absolute_url(theme):
            logger.info(
                f'Using theme "{theme}" from an absolute URL, no copy necessary',
            )
            return theme

        theme_path = None
        if not theme.endswith(".css"):
            assert default_theme_resource is not None
            with resources.as_file(
                default_theme_resource.joinpath(theme),
            ) as theme_path:
                theme_path = theme_path.with_suffix(".css").resolve(strict=True)
                logger.info(
                    f'Using built-in theme "{theme}" from "{theme_path.absolute()}"',
                )
        else:
            theme_path = Path(theme).resolve(strict=True)
            logger.info(f'Using theme "{theme_path.absolute()}"')

        theme_output_path = self.output_assets_path / theme_path.name
        self.__copy_to_output(theme_path, theme_output_path)

        relative_theme_path = theme_output_path.relative_to(
            file_using_theme_path.parent,
            walk_up=True,
        )

        return relative_theme_path

    def __copy_favicon(self, file_using_favicon_path: Path, favicon: str) -> Path | str:
        if self.__is_absolute_url(favicon):
            logger.info(
                f'Using favicon "{favicon}" from an absolute URL, no copy necessary',
            )
            return favicon

        favicon_path = Path(favicon).resolve(strict=True)
        logger.info(f'Using favicon "{favicon_path.absolute()}"')

        favicon_output_path = self.output_assets_path / favicon_path.name
        self.__copy_to_output(favicon_path, favicon_output_path)

        relative_favicon_path = favicon_output_path.relative_to(
            file_using_favicon_path.parent,
            walk_up=True,
        )

        return relative_favicon_path

    ################################################################################

    def __create_file(self, destination_path: Path, content: Any) -> None:
        if destination_path.exists():
            destination_path.write_text(content)
            logger.info(f'Overwritten: "{destination_path}"')
        else:
            destination_path.parent.mkdir(parents=True, exist_ok=True)
            destination_path.write_text(content)
            logger.info(f'Created file "{destination_path}"')

    def __copy_to_output(self, source_path: Path, destination_path: Path) -> Path:
        self.__copy(source_path, destination_path)
        relative_path = destination_path.relative_to(self.output_directory_path)
        return relative_path

    def __copy_to_output_relative_to_md_root(
        self,
        source_path: Path,
        md_root_path: Path,
    ) -> Path | None:
        source_path = source_path.resolve(strict=True)
        if not source_path.is_relative_to(md_root_path):
            logger.warning(
                f'"{source_path.absolute()}" is outside the markdown root directory, skipped!"',
            )
            return None

        relative_path = source_path.relative_to(md_root_path)
        destination_path = self.output_directory_path / relative_path

        self.__copy(source_path, destination_path)

        return relative_path

    def __copy(self, source_path, destination_path) -> None:
        if source_path.is_dir():
            self.__copy_tree(source_path, destination_path)
        else:
            self.__copy_file(source_path, destination_path)

    def __copy_tree(self, source_path, destination_path) -> None:
        overwrite = destination_path.exists()
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(source_path, destination_path, dirs_exist_ok=True)

        action = "Overwritten" if overwrite else "Copied"
        logger.info(
            f'{action} directory "{source_path.absolute()}" to "{destination_path.absolute()}"',
        )

    def __copy_file(self, source_path, destination_path) -> None:
        overwrite = destination_path.exists()
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(source_path, destination_path)

        action = "Overwritten" if overwrite else "Copied"
        logger.info(
            f'{action} file "{source_path.absolute()}" to "{destination_path.absolute()}"',
        )

    def __is_absolute_url(self, url: str) -> bool:
        return bool(urlparse(url).scheme)
